"""
Modules in this directory smooth over importing functionality that may
be present in different libraries, depending on the users'
system. Examples include astronomy utilities (astropy vs pyfits/pwcs)
and qt4 (available in PyQt4 and PySide)
"""
